# -*- coding: utf-8 -*-

import logging
import pathlib
import structlog
import traceback

from nplusone.core import exceptions


class Notifier(object):

    CONFIG_KEY = None
    ENABLED_DEFAULT = False

    @classmethod
    def is_enabled(cls, config):
        return (config.get(cls.CONFIG_KEY)
                or (cls.CONFIG_KEY not in config and cls.ENABLED_DEFAULT))

    def __init__(self, config):
        self.config = config  # pragma: no cover

    def notify(self, model, field):
        pass  # pragma: no cover


def get_relevant_spark_frames():
    stack = traceback.extract_stack()
    relevant_frames = [
        frame for frame in reversed(stack)
        if frame.filename.startswith(str(pathlib.Path().absolute()))
        and 'spark' in frame.filename
    ]
    return relevant_frames


class LogNotifier(Notifier):

    CONFIG_KEY = 'NPLUSONE_LOG'
    ENABLED_DEFAULT = True

    def __init__(self, config):
        self.logger = config.get('NPLUSONE_LOGGER', structlog.get_logger())
        self.level = config.get('NPLUSONE_LOG_LEVEL', logging.DEBUG)

        self.verbose = config.get('NPLUSONE_VERBOSE', False)
        log_func_map = {
            logging.INFO: self.logger.info,
            logging.WARN: self.logger.warn,
            logging.ERROR: self.logger.error,
            logging.DEBUG: self.logger.debug,
            logging.CRITICAL: self.logger.critical,
        }
        self.log_func = log_func_map.get(self.level, logging.INFO)

    def notify(self, message):
        stack = traceback.extract_stack()
        relevant_frames = [
            frame for frame in reversed(stack)
            if frame.filename.startswith(str(pathlib.Path().absolute()))
            and 'spark' in frame.filename
        ]
        relevant_frames = get_relevant_spark_frames()
        relevant_frame = relevant_frames[0]

        # This assumes we used structlog.get_logger to create our logger.
        log_info = {
            'filename': relevant_frame.filename,
            'line': relevant_frame.lineno,
            'name': relevant_frame.name,
        }

        if self.verbose:
            log_info['frames'] = '\n'.join([
                f'{frame.filename}, {frame.lineno}, {frame.name}'
                for frame in relevant_frames[1:]
            ])

        self.log_func(message.message, **log_info)


class ErrorNotifier(Notifier):

    CONFIG_KEY = 'NPLUSONE_RAISE'
    ENABLED_DEFAULT = False

    def __init__(self, config):
        self.error = config.get('NPLUSONE_ERROR', exceptions.NPlusOneError)

    def notify(self, message):
        stack = traceback.extract_stack()
        relevant_frame = get_relevant_spark_frames()[0]
        raise self.error(message.message + ', ' +
                         str(relevant_frame)[len('<FrameSummary '):-1])


def init(config):
    return [
        notifier(config) for notifier in (LogNotifier, ErrorNotifier)
        if notifier.is_enabled(config)
    ]



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from peewee import BaseQuery
from peewee import SelectQuery
from peewee import ForeignKeyAccessor
from peewee import BackrefAccessor
from peewee import ManyToManyQuery
from peewee import BaseModelSelect
from peewee import database_required

from nplusone.core import signals


def parse_get_object(args, kwargs, context):
    accessor, instance = args
    return accessor.field.model, to_key(instance), accessor.field.name


def parse_reverse_get(args, kwargs, context):
    accessor, instance = args
    return accessor.field.rel_model, to_key(instance), accessor.field.backref


def get_rel_instance(self, instance):
    value = instance.__data__.get(self.name)
    if value is not None or self.name in instance.__rel__:
        if self.name not in instance.__rel__:
            signals.lazy_load.send(
                signals.get_worker(),
                args=(self, instance),
                parser=parse_get_object,
            )
            obj = self.rel_model.get(self.field.rel_field == value)
            instance.__rel__[self.name] = obj
        return instance.__rel__[self.name]
    elif not self.field.null:
        raise self.rel_model.DoesNotExist
    return value
ForeignKeyAccessor.get_rel_instance = get_rel_instance


def backref_get(self, instance, instance_type=None):
    if instance is not None:
        dest = self.field.rel_field.name
        query = (self.rel_model
                .select()
                .where(self.field == getattr(instance, dest)))
        # Mark query with context so that we can emit a `lazyload` signal
        # if evaluated during `BaseQuery.execute`
        query._context = {
            'args': (self, instance),
            'kwargs': {'instance_type', instance_type},
        }
        return query
    return self  # pragma: no cover; pasted from peewee
BackrefAccessor.__get__ = backref_get


def to_key(instance):
    model = type(instance)
    return ':'.join([model.__name__, format(instance.get_id())])


def parse_load(args, kwargs, context, ret):
    return [to_key(row) for row in ret]


def is_single(offset, limit):
    return limit is not None and limit - (offset or 0) == 1


original_model_select_iter = BaseModelSelect.__iter__
def model_select_iter(self):
    if isinstance(self, ManyToManyQuery):
        signals.lazy_load.send(
            signals.get_worker(),
            args=(self._accessor, self._instance),
            parser=parse_get_object,
        )
    return original_model_select_iter(self)
BaseModelSelect.__iter__ = model_select_iter


original_query_execute = BaseQuery.execute
def query_execute(self, database):
    ret = original_query_execute(self, database)
    if hasattr(self, '_context'):
        # Query has been marked as lazy during backref lookup
        signals.lazy_load.send(
            signals.get_worker(),
            args=self._context['args'],
            kwargs=self._context['kwargs'],
            parser=parse_reverse_get,
        )
    if not isinstance(self, SelectQuery):
        return ret
    signal = (
        signals.ignore_load
        if is_single(self._offset, self._limit)
        else signals.load
    )
    signal.send(
        signals.get_worker(),
        args=(self, ),
        ret=list(ret),
        parser=parse_load,
    )
    return ret
BaseQuery.execute = database_required(query_execute)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import inspect
import itertools

from sqlalchemy.orm import query
from sqlalchemy.orm import loading
from sqlalchemy.orm import attributes
from sqlalchemy.orm import strategies

from nplusone.core import signals


def to_key(instance):
    model = type(instance)
    return ':'.join(
        [model.__name__] +
        [
            format(instance.__dict__.get(key.key))  # Avoid recursion on __get__
            for key in get_primary_keys(model)
        ]
    )


def get_primary_keys(model):
    mapper = model.__mapper__
    return [
        mapper.get_property_by_column(column)
        for column in mapper.primary_key
    ]


def parse_load(args, kwargs, context, ret):
    return [
        to_key(row) for row in ret
        if hasattr(row, '__table__')
    ]


def parse_lazy_load(args, kwargs, context):
    loader, state, _ = args
    return state.object.__class__, to_key(state.object), loader.parent_property.key


def parse_attribute_get(args, kwargs, context):
    attr, instance = args[:2]
    if instance is None:
        return None
    return attr.class_, attr.key, [to_key(instance)]


strategies.LazyLoader._load_for_state = signals.signalify(
    signals.lazy_load,
    strategies.LazyLoader._load_for_state,
    parser=parse_lazy_load,
)


def parse_populate(args, kwargs, context):
    query_context = args[0]
    state = args[2]
    instance = state.object
    return instance.__class__, context['key'], [to_key(instance)], id(query_context)


# Emit `eager_load` on populating from `joinedload` or `subqueryload`
original_populate_full = loading._populate_full
def _populate_full(*args, **kwargs):
    ret = original_populate_full(*args, **kwargs)
    context = inspect.getcallargs(original_populate_full, *args, **kwargs)
    for key, _ in context['populators'].get('eager', []):
        if context['dict_'].get(key):
            signals.eager_load.send(
                signals.get_worker(),
                args=args,
                kwargs=kwargs,
                context={'key': key},
                parser=parse_populate,
            )
    return ret
loading._populate_full = _populate_full


attributes.InstrumentedAttribute.__get__ = signals.signalify(
    signals.touch,
    attributes.InstrumentedAttribute.__get__,
    parser=parse_attribute_get,
)


def is_single(offset, limit):
    return limit is not None and limit - (offset or 0) == 1


original_query_iter = query.Query.__iter__
def query_iter(self):
    ret, clone = itertools.tee(original_query_iter(self))
    signal = (
        signals.ignore_load
        if is_single(self._offset, self._limit)
        else signals.load
    )
    signal.send(
        signals.get_worker(),
        args=(self, ),
        ret=list(clone),
        parser=parse_load,
    )
    return ret
query.Query.__iter__ = query_iter


def parse_get(args, kwargs, context, ret):
    return [to_key(ret)] if hasattr(ret, '__table__') else []


# Ignore records loaded during `one`
for method in ['one_or_none', 'one']:
    try:
        original = getattr(query.Query, method)
    except AttributeError:
        continue
    decorated = signals.signalify(signals.ignore_load, original, parse_get)
    setattr(query.Query, method, decorated)



# -*- coding: utf-8 -*-

import inspect

PATTERNS = ['site-packages', 'py.test', 'nplusone']


def get_caller(patterns=None):
    frames = inspect.stack()
    patterns = patterns or PATTERNS
    return next(
        (
            each for each in frames
            if each[4] and not any(pattern in each[1] for pattern in patterns)
        ),
        None,
    )



# -*- coding: utf-8 -*-

from nplusone.core import profiler


class NPlusOneMiddleware(object):

    def __init__(self, app, whitelist=None):
        self.app = app
        self.whitelist = whitelist

    def __call__(self, environ, start_response):
        with profiler.Profiler(whitelist=self.whitelist):
            return self.app(environ, start_response)



# -*- coding: utf-8 -*-

import six

from flask import g
from flask import request

from nplusone.core import signals
from nplusone.core import listeners
from nplusone.core import notifiers
import nplusone.ext.sqlalchemy  # noqa


def get_worker():
    try:
        return request._get_current_object()
    except RuntimeError:
        return None


def setup_state():
    signals.get_worker = get_worker
setup_state()


class NPlusOne(object):
    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def load_config(self, app):
        self.notifiers = notifiers.init(app.config)
        self.whitelist = [
            listeners.Rule(**item)
            for item in app.config.get('NPLUSONE_WHITELIST', [])
        ]

    def init_app(self, app):
        @app.before_request
        def connect():
            self.load_config(app)
            g.listeners = getattr(g, 'listeners', {})
            for name, listener_type in six.iteritems(listeners.listeners):
                g.listeners[name] = listener_type(self)
                g.listeners[name].setup()

        @app.after_request
        def disconnect(response):
            for name in six.iterkeys(listeners.listeners):
                listener = g.listeners.pop(name, None)
                if listener:
                    listener.teardown()
            return response

    def notify(self, message):
        if not message.match(self.whitelist):
            for notifier in self.notifiers:
                notifier.notify(message)

    def ignore(self, signal):
        return signals.ignore(getattr(signals, signal))



# -*- coding: utf-8 -*-

import six
import fnmatch
from collections import defaultdict

from nplusone.core import signals


class Rule(object):

    def __init__(self, label=None, model=None, field=None):
        self.label = label
        self.model = model
        self.field = field

    def compare(self, label, model, field):
        return (
            (self.label or self.model or self.field) and
            (self.label is None or self.label == label) and
            (self.model is None or self.match_model(model)) and
            (self.field is None or self.field == field)
        )

    def match_model(self, model):
        return (
            self.model is model or (
                isinstance(self.model, six.string_types) and
                fnmatch.fnmatch(model.__name__, self.model)
            )
        )


class Message(object):

    label = ''
    formatter = ''

    def __init__(self, model, field):
        self.model = model
        self.field = field

    @property
    def message(self):
        return self.formatter.format(
            label=self.label,
            model=self.model.__name__,
            field=self.field,
        )

    def match(self, rules):
        return any(
            rule.compare(self.label, self.model, self.field)
            for rule in rules
        )


class LazyLoadMessage(Message):
    label = 'n_plus_one'
    formatter = 'Potential n+1 query detected on `{model}.{field}`'


class EagerLoadMessage(Message):
    label = 'unused_eager_load'
    formatter = 'Potential unnecessary eager load detected on `{model}.{field}`'


class Listener(object):

    def __init__(self, parent):
        self.parent = parent

    def setup(self):
        pass  # pragma: no cover

    def teardown(self):
        pass  # pragma: no cover


class LazyListener(Listener):

    def setup(self):
        self.loaded, self.ignore = set(), set()
        signals.load.connect(self.handle_load, sender=signals.get_worker())
        signals.ignore_load.connect(self.handle_ignore, sender=signals.get_worker())
        signals.lazy_load.connect(self.handle_lazy, sender=signals.get_worker())

    def handle_load(self, caller, args=None, kwargs=None, context=None, ret=None,
                    parser=None):
        instances = parser(args, kwargs, context, ret)
        self.loaded.update(instances)

    def handle_ignore(self, caller, args=None, kwargs=None, context=None, ret=None,
                      parser=None):
        instances = parser(args, kwargs, context, ret)
        self.ignore.update(instances)

    def handle_lazy(self, caller, args=None, kwargs=None, context=None, ret=None,
                    parser=None):
        model, instance, field = parser(args, kwargs, context)
        self.ignore = set() # TODO: Better configure what we want to ignore
        if instance in self.loaded and instance not in self.ignore:
            message = LazyLoadMessage(model, field)
            self.parent.notify(message)


class EagerListener(Listener):

    def setup(self):
        signals.eager_load.connect(self.handle_eager, sender=signals.get_worker())
        self.tracker = EagerTracker()
        self.touched = []

    def teardown(self):
        self.log_eager()

    def handle_eager(self, caller, args=None, kwargs=None, context=None, ret=None,
                     parser=None):
        self.tracker.track(*parser(args, kwargs, context))
        signals.touch.connect(self.handle_touch, sender=signals.get_worker())

    def handle_touch(self, caller, args=None, kwargs=None, context=None, ret=None,
                     parser=None):
        self.touched.append(parser(args, kwargs, context))

    def log_eager(self):
        self.tracker.prune([each for each in self.touched if each])
        for model, field in self.tracker.unused:
            message = EagerLoadMessage(model, field)
            self.parent.notify(message)


class EagerTracker(object):
    """Data structure for tracking eager-loaded and subsequently touched
    related rows. Eager-loaded rows are stored in a dict mapping associations
    to nested dicts mapping query keys to instances.
    """
    def __init__(self):
        self.data = defaultdict(lambda: defaultdict(set))

    def track(self, model, field, instances, key):
        self.data[(model, field)][key].update(instances)

    def prune(self, touched):
        for model, field, touch_instances in touched:
            group = self.data[(model, field)]
            for key, fetch_instances in list(group.items()):
                if touch_instances and fetch_instances.intersection(touch_instances):
                    group.pop(key, None)

    @property
    def unused(self):
        return [
            (model, field)
            for (model, field), group in self.data.items()
            if group
        ]


listeners = {
    'lazy_load': LazyListener,
    'eager_load': EagerListener,
}



# -*- coding: utf-8 -*-

import fnmatch
import weakref

import six

from django.conf import settings

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object

from nplusone.core import listeners
from nplusone.core import notifiers


class DjangoRule(listeners.Rule):

    def match_model(self, model):
        return (
            self.model is model or (
                isinstance(self.model, six.string_types) and
                fnmatch.fnmatch(
                    '{0}.{1}'.format(model._meta.app_label, model.__name__),
                    self.model,
                )
            )
        )


class NPlusOneMiddleware(MiddlewareMixin):

    def __init__(self, *args, **kwargs):
        super(NPlusOneMiddleware, self).__init__(*args, **kwargs)
        self.listeners = weakref.WeakKeyDictionary()

    def load_config(self):
        self.notifiers = notifiers.init(vars(settings._wrapped))
        self.whitelist = [
            DjangoRule(**item)
            for item in getattr(settings, 'NPLUSONE_WHITELIST', [])
        ]

    def process_request(self, request):
        self.load_config()
        self.listeners[request] = self.listeners.get(request, {})
        for name, listener_type in six.iteritems(listeners.listeners):
            self.listeners[request][name] = listener_type(self)
            self.listeners[request][name].setup()

    def process_response(self, request, response):
        for name, listener_type in six.iteritems(listeners.listeners):
            listener = self.listeners.get(request, {}).pop(name, None)
            if listener:
                listener.teardown()
        return response

    def notify(self, message):
        if not message.match(self.whitelist):
            for notifier in self.notifiers:
                notifier.notify(message)



# -*- coding: utf-8 -*-

import copy
import inspect
import functools
import importlib
import threading

import django
from django.db.models import query
from django.db.models import Model

from nplusone.core import signals

if django.VERSION >= (1, 9):  # pragma: no cover
    from django.db.models.fields.related_descriptors import (
        ReverseOneToOneDescriptor,
        ForwardManyToOneDescriptor,
        create_reverse_many_to_one_manager,
        create_forward_many_to_many_manager,
    )
else:  # pragma: no cover
    from django.db.models.fields.related import (
        SingleRelatedObjectDescriptor as ReverseOneToOneDescriptor,
        ReverseSingleRelatedObjectDescriptor as ForwardManyToOneDescriptor,
        create_foreign_related_manager as create_reverse_many_to_one_manager,
        create_many_related_manager as create_forward_many_to_many_manager,
    )


def get_worker():
    return str(threading.current_thread().ident)


def setup_state():
    signals.get_worker = get_worker
setup_state()


def to_key(instance):
    model = type(instance)
    return ':'.join([model.__name__, format(instance.pk)])


def patch(original, patched):
    module = importlib.import_module(original.__module__)
    setattr(module, original.__name__, patched)


def signalify_queryset(func, parser=None, **context):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        queryset = func(*args, **kwargs)
        ctx = copy.copy(context)
        ctx['args'] = context.get('args', args)
        ctx['kwargs'] = context.get('kwargs', kwargs)
        queryset._clone = signalify_queryset(queryset._clone, parser=parser, **ctx)
        queryset._fetch_all = signalify_fetch_all(queryset, parser=parser, **ctx)
        queryset._context = ctx
        return queryset
    return wrapped


def signalify_fetch_all(queryset, parser=None, **context):
    """Signal lazy load when `QuerySet._fetch_all` fetches rows. Note: patch
    `_fetch_all` instead of `iterator` since, as of Django 1.11, the former is
    used for all fetches while the latter is not.
    """
    func = queryset._fetch_all
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        if queryset._result_cache is None:
            signals.lazy_load.send(
                get_worker(),
                args=args,
                kwargs=kwargs,
                ret=None,
                context=context,
                parser=parser,
            )
        return func(*args, **kwargs)
    return wrapped


def get_related_name(model):
    return '{0}_set'.format(model._meta.model_name)


def parse_field(field):
    return (
        (
            field.related_model  # Django >= 1.8
            if hasattr(field, 'related_model')
            else field.related_field.model  # Django <= 1.8
        ),
        (
            field.remote_field.name  # Django >= 1.8
            if hasattr(field, 'remote_field')
            else field.rel.related_name  # Django <= 1.8
        ) or get_related_name(field.related_model),
    )


def parse_reverse_field(field):
    return field.model, field.name


def parse_related(context):
    if 'rel' in context:  # pragma: no cover
        rel = context['rel']
        return parse_related_parts(rel.model, rel.related_name, rel.related_model)
    else:  # pragma: no cover
        field = context['rel_field']
        model = field.related_field.model
        related_name = field.rel.related_name
        related_model = context['rel_model']
        return parse_related_parts(model, related_name, related_model)


def parse_related_parts(model, related_name, related_model):
    return (
        model,
        related_name or get_related_name(related_model),
    )


def parse_reverse_one_to_one_queryset(args, kwargs, context):
    descriptor = context['args'][0]
    field = descriptor.related.field
    model, name = parse_field(field)
    instance = context['kwargs']['instance']
    return model, to_key(instance), name


def parse_forward_many_to_one_queryset(args, kwargs, context):
    descriptor = context['args'][0]
    instance = context['kwargs']['instance']
    return descriptor.field.model, to_key(instance), descriptor.field.name


def parse_many_related_queryset(args, kwargs, context):
    rel = context['rel']
    manager = context['args'][0]
    model = manager.instance.__class__
    related_model = (
        manager.target_field.related_model  # Django >= 1.8
        if hasattr(manager.target_field, 'related_model')
        else manager.target_field.related_field.model  # Django <= 1.8
    )
    field = manager.prefetch_cache_name if rel.related_name else None
    return (
        model,
        to_key(manager.instance),
        field or get_related_name(related_model),
    )


def parse_foreign_related_queryset(args, kwargs, context):
    model, name = parse_related(context)
    descriptor = context['args'][0]
    return model, to_key(descriptor.instance), name


query.prefetch_one_level = signals.designalify(
    signals.lazy_load,
    query.prefetch_one_level,
)


def parse_get(args, kwargs, context, ret):
    return [to_key(ret)] if isinstance(ret, Model) else []


# Ignore records loaded during `get`
query.QuerySet.get = signals.signalify(
    signals.ignore_load,
    query.QuerySet.get,
    parser=parse_get,
)


ReverseOneToOneDescriptor.get_queryset = signalify_queryset(
    ReverseOneToOneDescriptor.get_queryset,
    parser=parse_reverse_one_to_one_queryset,
)
ForwardManyToOneDescriptor.get_queryset = signalify_queryset(
    ForwardManyToOneDescriptor.get_queryset,
    parser=parse_forward_many_to_one_queryset,
)


def _create_forward_many_to_many_manager(*args, **kwargs):
    context = inspect.getcallargs(create_forward_many_to_many_manager, *args, **kwargs)
    manager = create_forward_many_to_many_manager(*args, **kwargs)
    manager.get_queryset = signalify_queryset(
        manager.get_queryset,
        parser=parse_many_related_queryset,
        **context
    )
    return manager
patch(create_forward_many_to_many_manager, _create_forward_many_to_many_manager)


def _create_reverse_many_to_one_manager(*args, **kwargs):
    context = inspect.getcallargs(create_reverse_many_to_one_manager, *args, **kwargs)
    manager = create_reverse_many_to_one_manager(*args, **kwargs)

    manager.get_queryset = signalify_queryset(
        manager.get_queryset,
        parser=parse_foreign_related_queryset,
        **context
    )
    return manager
patch(create_reverse_many_to_one_manager, _create_reverse_many_to_one_manager)


def parse_forward_many_to_one_get(args, kwargs, context):
    descriptor, instance, _ = args
    if instance is None:
        return None
    field, model = parse_reverse_field(descriptor.field)
    return field, model, [to_key(instance)]


ForwardManyToOneDescriptor.__get__ = signals.signalify(
    signals.touch,
    ForwardManyToOneDescriptor.__get__,
    parser=parse_forward_many_to_one_get,
)


def parse_reverse_one_to_one_get(args, kwargs, context):
    descriptor, instance = args[:2]
    if instance is None:
        return None
    model, field = parse_field(descriptor.related.field)
    return model, field, [to_key(instance)]


ReverseOneToOneDescriptor.__get__ = signals.signalify(
    signals.touch,
    ReverseOneToOneDescriptor.__get__,
    parser=parse_reverse_one_to_one_get,
)


def parse_fetch_all(args, kwargs, context):
    self = args[0]
    if hasattr(self, '_context'):
        manager = self._context['args'][0]
        instance = manager.instance
        # Handle iteration over many-to-many relationship
        if manager.__class__.__name__ == 'ManyRelatedManager':
            return (
                instance.__class__,
                parse_manager_field(manager, self._context['rel']),
                [to_key(instance)],
            )
        # Handle iteration over one-to-many relationship
        else:
            model, field = parse_related(self._context)
            return model, field, [to_key(instance)]


def parse_manager_field(manager, rel):
    if manager.reverse:
        return rel.related_name or get_related_name(rel.related_model)
    return rel.field.name or get_related_name(rel.model)


def parse_load(args, kwargs, context, ret):
    return [
        to_key(row)
        for row in ret
        if isinstance(row, Model)
    ]


def is_single(low, high):
    return high is not None and high - low == 1


# On queryset fetch, emit `touch` if results have been prefetched; emit `load`
# if the query requests more than one record, else `ignore_load`. Note: we patch
# `_fetch_all` rather than `__iter__` to handle iteration over empty querysets
# in Django templates, which does not call `__iter__`.
original_fetch_all = query.QuerySet._fetch_all
def fetch_all(self):
    if self._prefetch_done:
        signals.touch.send(
            get_worker(),
            args=(self, ),
            parser=parse_fetch_all,
        )
    original_fetch_all(self)
    signal = (
        signals.ignore_load
        if is_single(self.query.low_mark, self.query.high_mark)
        else signals.load
    )
    signal.send(
        get_worker(),
        args=(self, ),
        ret=self._result_cache,
        parser=parse_load,
    )
query.QuerySet._fetch_all = fetch_all


original_related_populator_init = query.RelatedPopulator.__init__
def related_populator_init(self, *args, **kwargs):
    original_related_populator_init(self, *args, **kwargs)
    self.__nplusone__ = {
        'args': args,
        'kwargs': kwargs,
    }
query.RelatedPopulator.__init__ = related_populator_init


def parse_eager_select(args, kwargs, context):
    populator = args[0]
    instance = args[2]
    meta = populator.__nplusone__
    klass_info, select, _ = meta['args']
    field = klass_info['field']
    model, name = (
        parse_field(field)
        if instance._meta.model != field.model
        else parse_reverse_field(field)
    )
    return model, name, [to_key(instance)], id(select)


# Emit `eager_load` on populating from `select_related`
query.RelatedPopulator.populate = signals.signalify(
    signals.eager_load,
    query.RelatedPopulator.populate,
    parser=parse_eager_select,
)


def parse_eager_join(args, kwargs, context):
    instances, descriptor, fetcher, level = args
    model = instances[0].__class__
    field, _ = fetcher.get_current_to_attr(level)
    keys = [to_key(instance) for instance in instances]
    return model, field, keys, id(instances)


# Emit `eager_load` on populating from `prefetch_related`
query.prefetch_one_level = signals.signalify(
    signals.eager_load,
    query.prefetch_one_level,
    parser=parse_eager_join,
)


# Emit `touch` on indexing into prefetched `QuerySet` instances
original_getitem_queryset = query.QuerySet.__getitem__
def getitem_queryset(self, index):
    if self._prefetch_done:
        signals.touch.send(
            get_worker(),
            args=(self, ),
            parser=parse_fetch_all,
        )
    return original_getitem_queryset(self, index)
query.QuerySet.__getitem__ = getitem_queryset



# -*- coding: utf-8 -*-

import functools
import contextlib

import blinker


load = blinker.Signal()
ignore_load = blinker.Signal()
lazy_load = blinker.Signal()
eager_load = blinker.Signal()
touch = blinker.Signal()


def get_worker(*args, **kwargs):
    return blinker.ANY


def signalify(signal, func, parser=None, **context):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        ret = func(*args, **kwargs)
        signal.send(
            get_worker(),
            args=args,
            kwargs=kwargs,
            ret=ret,
            context=context,
            parser=parser,
        )
        return ret
    return wrapped


def designalify(signal, func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        with ignore(signal):
            return func(*args, **kwargs)
    return wrapped


@contextlib.contextmanager
def ignore(signal, sender=None):
    sender = sender or get_worker()
    receivers = list(signal.receivers_for(sender))
    for receiver in receivers:
        signal.disconnect(receiver, sender=sender)
    try:
        yield
    finally:
        for receiver in receivers:
            signal.connect(receiver, sender=sender)



# -*- coding: utf-8 -*-

import six

from nplusone.core import listeners
from nplusone.core import exceptions


class Profiler(object):

    def __init__(self, whitelist=None):
        self.whitelist = [
            listeners.Rule(**item)
            for item in (whitelist or [])
        ]

    def __enter__(self):
        self.listeners = {}
        for name, listener_type in six.iteritems(listeners.listeners):
            self.listeners[name] = listener_type(self)
            self.listeners[name].setup()

    def __exit__(self, *exc):
        for name in six.iterkeys(listeners.listeners):
            self.listeners.pop(name).teardown()

    def notify(self, message):
        if not message.match(self.whitelist):
            raise exceptions.NPlusOneError(message.message)
